#!/usr/bin/env python3
# -*- coding: utf-8; -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import os
import shutil

import params.src.indices as indices
  
def evaluate(parameters, surrogate_parameters, images_parameters):
    import imp
    model = imp.load_source('Metafor_call',parameters['Metafor_model']+'/model.py')
    ## Read the parameters
    fname = surrogate_parameters['fname']
    d = surrogate_parameters['d']
    M = surrogate_parameters['M']
    N = surrogate_parameters['N']

    Dom_min = surrogate_parameters['Dom_min']
    Dom_max = surrogate_parameters['Dom_max']
    P_name = surrogate_parameters['P_name']
    solver = surrogate_parameters['solver']

    normalisation = surrogate_parameters['normalisation']
    ##
    if surrogate_parameters['function_type'] == 'monomial': 
        from params.src.monomials import phi   
        from params.src.monomials import compute_n_max
        n_max = compute_n_max(N,d)
    elif surrogate_parameters['function_type'] == 'monomial_1norm': 
        from params.src.monomials_1norm import phi   
        from params.src.monomials_1norm import compute_n_max
        n_max = compute_n_max(N,d)
    if surrogate_parameters['training_points'] == 'gauss': 
        import params.src.gauss as tp
    elif surrogate_parameters['training_points'] == 'gamma': 
        import params.arnst_ponthot.src.gamma as tp

    ##

    X, w = tp.grid(M)  

    m_max = 1
    for j in range(0,d):
        m_max = m_max*M[j] 

    i =      np.zeros((d,), dtype =np.int)
    i_zero = np.zeros((d,), dtype =np.int)
    i_one =  np.ones((d,), dtype =np.int)
    p = 0

    xi = np.zeros((d,))

    if surrogate_parameters['display']: 
        import params.src.plot as plot
        fig, ax, current_title = plot.init2(d, images_parameters)

        if not os.path.exists('images'):
            os.mkdir('images')
        else:
            shutil.rmtree('images')
            os.mkdir('images')

    m_i = 0
    sol = np.zeros((m_max,1))

    Z = np.zeros((m_max,n_max))
    W_sr = np.zeros((m_max,m_max))    

    while np.all(i != -i_one):   


        if normalisation == True:
            for j in range(0,d):
                xi[j] = X[i[j],j]
        elif normalisation == False:
            for j in range(0,d):
                xi[j] = X[i[j],j]
                xi[j] = Dom_min[j] + (Dom_max[j] - Dom_min[j])*(xi[j]+1.)/2.
                parameters[P_name[j]] = xi[j]

        # -------------------

        Z[m_i,:] = np.transpose(phi(xi,N,n_max,d))

        tmp = 1
        for j in range(0,d):
            tmp = tmp*w[i[j],j]

        W_sr[m_i,m_i]  = np.sqrt(tmp)   


        # -------------------

        if normalisation == True:
            for j in range(0,d):
                xi[j] = Dom_min[j] + (Dom_max[j] - Dom_min[j])*(xi[j]+1.)/2.
                parameters[P_name[j]] = xi[j]

        # -------------------

        name = 'i_'
        for j in range(0,d-1):
            name = name + str(i[j]) + '_'
        name = name + str(i[d-1]) + '_'
        name = name + 'M_' 
        for j in range(0,d-1):
            name = name + str(M[j]) + '_'
        name = name + str(M[d-1]) 

        # -------------------

        if os.path.exists(name):
            shutil.rmtree(name)
        os.mkdir(name)
        os.chdir(name)

        sol[m_i] = model.model(parameters)

        os.chdir('..')

        if surrogate_parameters['display']: 
            plot.scatter(xi,sol[m_i],d,fig,ax, True,m_i)
            

        m_i = m_i + 1 
       
        p,i = indices.my_next(p,i,M,d)


    #------------------
    Image = {}
    if surrogate_parameters['display']:
        Image['fig'] = fig
        Image['ax'] = ax
        Image['current_title'] = current_title
    return sol, Z, W_sr, Image


def Z_evaluate(parameters, surrogate_parameters):
    import imp
    model = imp.load_source('Metafor_call',parameters['Metafor_model']+'/model.py')
    ## Read the parameters
    fname = surrogate_parameters['fname']
    d = surrogate_parameters['d']
    M = surrogate_parameters['M']
    N = surrogate_parameters['N']

    Dom_min = surrogate_parameters['Dom_min']
    Dom_max = surrogate_parameters['Dom_max']
    P_name = surrogate_parameters['P_name']
    solver = surrogate_parameters['solver']

    normalisation = surrogate_parameters['normalisation']
    ##
    if surrogate_parameters['function_type'] == 'monomial': 
        from params.src.monomials import phi   
        from params.src.monomials import compute_n_max
        n_max = compute_n_max(N,d)
    elif surrogate_parameters['function_type'] == 'monomial_1norm': 
        from params.src.monomials_1norm import phi   
        from params.src.monomials_1norm import compute_n_max
        n_max = compute_n_max(N,d)
    if surrogate_parameters['training_points'] == 'gauss': 
        import params.src.gauss as tp
    elif surrogate_parameters['training_points'] == 'gamma': 
        import params.arnst_ponthot.src.gamma as tp

    ##

    if surrogate_parameters['training_points'] != 'precomputed': 

        X, w = tp.grid(M)  

        m_max = 1
        for j in range(0,d):
            m_max = m_max*M[j] 
    if surrogate_parameters['training_points'] == 'precomputed': 


        import codecs, json 

        obj_text = codecs.open(surrogate_parameters['precomputed_training_points'], 'r', encoding='utf-8').read()
        b_new = json.loads(obj_text)
        X = np.array(b_new)


        #X = np.loadtxt('/home/kliegeois/Codes/waves/params/CVT/python/cvt_circle7.txt')
        tmp = np.shape(X)
        m_max = tmp[0] 

        obj_text = codecs.open(surrogate_parameters['precomputed_weights'], 'r', encoding='utf-8').read()
        b_new = json.loads(obj_text)
        w = np.array(b_new)

    i =      np.zeros((d,), dtype =np.int)
    i_zero = np.zeros((d,), dtype =np.int)
    i_one =  np.ones((d,), dtype =np.int)
    p = 0

    xi = np.zeros((d,))

    m_i = 0
    sol = np.zeros((m_max,1))

    Z = np.zeros((m_max,n_max))
    W_sr = np.zeros((m_max,m_max))    

    Xi = np.zeros((m_max,d))
    Xi2 = np.zeros((m_max,d))
    while np.all(i != -i_one):   


        if surrogate_parameters['training_points'] == 'precomputed':
            means = surrogate_parameters['means']
            variances = surrogate_parameters['variances']
            correlations = surrogate_parameters['correlations']

            for j in range(0,d):
                xi[j] = ((X[m_i,j]-means[j])/(surrogate_parameters['nstddeviation']*np.sqrt(variances[j])))

        elif surrogate_parameters['training_points'] == 'gamma': 
            means = surrogate_parameters['means']
            variances = surrogate_parameters['variances']
            correlations = surrogate_parameters['correlations']

            xi_in = np.zeros((d,1))
            for j in range(0,d):
                xi_in[j] = X[i[j],j]
            xi = tp.f_mapping(xi_in,d,means,variances,correlations,surrogate_parameters['f_mapping_method'])

            #print xi

        elif normalisation == True:
            for j in range(0,d):
                xi[j] = X[i[j],j]
        elif normalisation == False:
            for j in range(0,d):
                xi[j] = X[i[j],j]
                xi[j] = Dom_min[j] + (Dom_max[j] - Dom_min[j])*(xi[j]+1.)/2.
                parameters[P_name[j]] = xi[j]

        # -------------------
        xi2 = np.zeros((d,))
        if surrogate_parameters['training_points'] == 'gamma': 
            
            xi2[0] = (xi[0][0]-means[0]) / (surrogate_parameters['nstddeviation']*np.sqrt(variances[0]))
            xi2[1] = (xi[1][0]-means[1]) / (surrogate_parameters['nstddeviation']*np.sqrt(variances[1]))    
            Z[m_i,:] = np.transpose(phi(xi2,N,n_max,d))
            #print xi2
        else:
            Z[m_i,:] = np.transpose(phi(xi,N,n_max,d))

        tmp = 1
        if surrogate_parameters['training_points'] != 'precomputed':
            for j in range(0,d):
                tmp = tmp*w[i[j],j]
        else:
            tmp = w[m_i]

        W_sr[m_i,m_i]  = np.sqrt(tmp)     



        Xi[m_i,:] = xi.T
        Xi2[m_i,:] = xi2.T
        W_sr[m_i,m_i]  = np.sqrt(tmp)  
            
        m_i = m_i + 1 
       
        p,i = indices.my_next(p,i,M,d)

    return Z, W_sr, Xi, Xi2


def scatterplot(sol, surrogate_parameters, images_parameters):
    Image = {}
    if surrogate_parameters['display']: 

        d = surrogate_parameters['d']
        M = surrogate_parameters['M']

        Dom_min = surrogate_parameters['Dom_min']
        Dom_max = surrogate_parameters['Dom_max']

        if surrogate_parameters['training_points'] == 'gauss': 
            import params.src.gauss as tp
        elif surrogate_parameters['training_points'] == 'gamma': 
            import params.arnst_ponthot.src.gamma as tp

        import params.src.plot as plot
        fig, ax, current_title = plot.init2(d, images_parameters)

        if not os.path.exists('images'):
            os.mkdir('images')
        else:
            shutil.rmtree('images')
            os.mkdir('images')

        if surrogate_parameters['training_points'] != 'precomputed': 

            X, w = tp.grid(M)  

            m_max = 1
            for j in range(0,d):
                m_max = m_max*M[j] 
        if surrogate_parameters['training_points'] == 'precomputed': 


            import codecs, json 

            obj_text = codecs.open(surrogate_parameters['precomputed_training_points'], 'r', encoding='utf-8').read()
            b_new = json.loads(obj_text)
            X = np.array(b_new)

            tmp = np.shape(X)
            print(tmp)
            m_max = tmp[0] 
            print(max(X[:,0]))
            print(min(X[:,0]))
            print(max(X[:,1]))
            print(min(X[:,1]))


        xi = np.zeros((d,))
        xi_tp = np.zeros((m_max,d))    
        m_i = 0

        i =      np.zeros((d,), dtype =np.int)
        i_zero = np.zeros((d,), dtype =np.int)
        i_one =  np.ones((d,), dtype =np.int)
        p = 0
        
        while np.all(i != -i_one):  
            if surrogate_parameters['training_points'] == 'gamma': 
                means = surrogate_parameters['means']
                variances = surrogate_parameters['variances']
                correlations = surrogate_parameters['correlations']

                xi_in = np.zeros((d,1))
                for j in range(0,d):
                    xi_in[j] = X[i[j],j]
                xi = tp.f_mapping(xi_in,d,means,variances,correlations,surrogate_parameters['f_mapping_method']) 
            
            elif surrogate_parameters['training_points'] == 'precomputed': 
                for j in range(0,d):
                    xi[j] = X[m_i,j]

            else:
                for j in range(0,d):
                    xi[j] = X[i[j],j]
                    xi[j] = Dom_min[j] + (Dom_max[j] - Dom_min[j])*(xi[j]+1.)/2.
            for j in range(0,d):
                xi_tp[m_i,j] = xi[j]
            m_i = m_i + 1
            p,i = indices.my_next(p,i,M,d)

        plot.scatters(xi_tp,sol,d,fig,ax, True,m_i)
        Image['fig'] = fig
        Image['ax'] = ax
        Image['current_title'] = current_title
    return Image
