#!/usr/bin/env python3
# -*- coding: utf-8; -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import os
import shutil

import params.src.indices as indices
  

def surrogate_creation(sol, Z, W_sr, surrogate_parameters, Image, images_parameters):
    if surrogate_parameters['display']: 
        import params.src.plot as plot
        fig = Image['fig']
        ax = Image['ax'] 
        current_title = Image['current_title']
    d = surrogate_parameters['d']
    M = surrogate_parameters['M']
    N = surrogate_parameters['N']

    m1 = surrogate_parameters['m1']
    m2 = surrogate_parameters['m2']

    Dom_min = surrogate_parameters['Dom_min']
    Dom_max = surrogate_parameters['Dom_max']
    P_name = surrogate_parameters['P_name']
    solver = surrogate_parameters['solver']

    normalisation = surrogate_parameters['normalisation']
    ##
    if surrogate_parameters['function_type'] == 'monomial': 
        from params.src.monomials import phi   
    elif surrogate_parameters['function_type'] == 'monomial_1norm': 
        from params.src.monomials_1norm import phi     
    ##

    shape = Z.shape
    m_max = shape[0]
    n_max = shape[1]

    x = np.linspace(-1.,1.,m1)
    y = np.linspace(-1.,1.,m2)
    X = np.zeros((m1,m2))
    Y = np.zeros((m1,m2))
    surrogate = np.zeros((m1,m2))
    X, Y = np.meshgrid(x, y)
    phi_tmp = np.zeros((m1,m2,n_max))

    #------------------
    
    fixed_value = np.zeros((d-2,))
    if d > 2:
        fixed_value = surrogate_parameters['fixed_value'] 
    for i in range(0,m1):
        for j in range(0,m2):
            xi = np.zeros((d,))

            if surrogate_parameters['training_points'] == 'gamma': 
                X[i,j] = Dom_min[0] + (Dom_max[0] - Dom_min[0])*(X[i,j]+1.)/2.
                Y[i,j] = Dom_min[1] + (Dom_max[1] - Dom_min[1])*(Y[i,j]+1.)/2.
                means = surrogate_parameters['means']
                variances = surrogate_parameters['variances']
               
                xi[0] = (X[i,j]-means[0])/(surrogate_parameters['nstddeviation']*np.sqrt(variances[0]))
                xi[1] = (Y[i,j]-means[1])/(surrogate_parameters['nstddeviation']*np.sqrt(variances[1]))
                for k in range(2,d):
                    xi[k] = (fixed_value[k-2]-means[k]) / np.sqrt(variances[k])


            elif surrogate_parameters['training_points'] == 'precomputed': 
                X[i,j] = Dom_min[0] + (Dom_max[0] - Dom_min[0])*(X[i,j]+1.)/2.
                Y[i,j] = Dom_min[1] + (Dom_max[1] - Dom_min[1])*(Y[i,j]+1.)/2.
                means = surrogate_parameters['means']
                variances = surrogate_parameters['variances']
               
                xi[0] = (X[i,j]-means[0]) / (surrogate_parameters['nstddeviation']*np.sqrt(variances[0]))
                xi[1] = (Y[i,j]-means[1]) / (surrogate_parameters['nstddeviation']*np.sqrt(variances[1]))
                for k in range(2,d):
                    xi[k] = (fixed_value[k-2]-means[k]) / np.sqrt(variances[k])

            elif normalisation == False:
                X[i,j] = Dom_min[0] + (Dom_max[0] - Dom_min[0])*(X[i,j]+1.)/2.
                Y[i,j] = Dom_min[1] + (Dom_max[1] - Dom_min[1])*(Y[i,j]+1.)/2.
            
                xi[0] = X[i,j]
                xi[1] = Y[i,j]
            else:
                xi[0] = X[i,j]
                xi[1] = Y[i,j]


            phi_tmp[i,j,:] = phi(xi,N,n_max,d)


    if normalisation == True:
        for i in range(0,m1):
            for j in range(0,m2):
                X[i,j] = Dom_min[0] + (Dom_max[0] - Dom_min[0])*(X[i,j]+1.)/2.
                Y[i,j] = Dom_min[1] + (Dom_max[1] - Dom_min[1])*(Y[i,j]+1.)/2.


    if solver == 'QR': 
        W_srZ = np.dot(W_sr,Z)
        print(np.linalg.cond(W_srZ))
        Q, R = np.linalg.qr(W_srZ)
        rhs = np.dot(np.transpose(Q),np.dot(W_sr,sol))
        s_hat = np.linalg.solve(R, rhs)
        for i in range(0,m1):
            for j in range(0,m2):
                surrogate[i,j] = np.dot(np.transpose(phi_tmp[i,j,:]),s_hat)


    if solver == 'sp': 
        W_srZ = np.dot(W_sr,Z)
        ZWZ = np.dot(np.transpose(W_srZ),W_srZ)
        print(np.linalg.cond(ZWZ))
        rhs = np.dot(np.transpose(W_srZ),np.dot(W_sr,sol))
        import scipy.sparse.linalg
        s_hat = scipy.sparse.linalg.spsolve(ZWZ, rhs, 'NATURAL', False)
        for i in range(0,m1):
            for j in range(0,m2):
                surrogate[i,j] = np.dot(np.transpose(phi_tmp[i,j,:]),s_hat)

    if solver == 'numpy_linalg': 
        W_srZ = np.dot(W_sr,Z)
        ZWZ = np.dot(np.transpose(W_srZ),W_srZ)
        print(np.linalg.cond(ZWZ))
        rhs = np.dot(np.transpose(W_srZ),np.dot(W_sr,sol))
        s_hat = np.linalg.solve(ZWZ, rhs)
        for i in range(0,m1):
            for j in range(0,m2):
                surrogate[i,j] = np.dot(np.transpose(phi_tmp[i,j,:]),s_hat)

    if solver == 'scipy_linalg': 
        W_srZ = np.dot(W_sr,Z)
        ZWZ = np.dot(np.transpose(W_srZ),W_srZ)
        print(np.linalg.cond(ZWZ))
        rhs = np.dot(np.transpose(W_srZ),np.dot(W_sr,sol))
        import scipy.sparse.linalg
        s_hat = scipy.linalg.solve(ZWZ, rhs)
        for i in range(0,m1):
            for j in range(0,m2):
                surrogate[i,j] = np.dot(np.transpose(phi_tmp[i,j,:]),s_hat)

    if solver == 'scipy_lu': 
        W_srZ = np.dot(W_sr,Z)
        ZWZ = np.dot(np.transpose(W_srZ),W_srZ)
        print(np.linalg.cond(ZWZ))
        rhs = np.dot(np.transpose(W_srZ),np.dot(W_sr,sol))
        import scipy.sparse.linalg
        PL, U = scipy.linalg.lu(ZWZ,True)
        print(np.linalg.cond(PL))
        print(np.linalg.cond(U))
        tmp = scipy.linalg.solve(PL,rhs)
        s_hat = scipy.linalg.solve(U,tmp)
        for i in range(0,m1):
            for j in range(0,m2):
                surrogate[i,j] = np.dot(np.transpose(phi_tmp[i,j,:]),s_hat)

    if solver == 'minres': 
        W_srZ = np.dot(W_sr,Z)
        ZWZ = np.dot(np.transpose(W_srZ),W_srZ)
        print(np.linalg.cond(ZWZ))
        rhs = np.dot(np.transpose(W_srZ),np.dot(W_sr,sol))
        import scipy.sparse.linalg
        s_hat, info = scipy.sparse.linalg.minres(ZWZ, rhs)
        for i in range(0,m1):
            for j in range(0,m2):
                surrogate[i,j] = np.dot(np.transpose(phi_tmp[i,j,:]),s_hat)

    if solver == 'gmres': 
        W_srZ = np.dot(W_sr,Z)
        ZWZ = np.dot(np.transpose(W_srZ),W_srZ)
        print(np.linalg.cond(ZWZ))
        rhs = np.dot(np.transpose(W_srZ),np.dot(W_sr,sol))
        import scipy.sparse.linalg
        s_hat, info = scipy.sparse.linalg.gmres(ZWZ, rhs)
        for i in range(0,m1):
            for j in range(0,m2):
                surrogate[i,j] = np.dot(np.transpose(phi_tmp[i,j,:]),s_hat)

    if solver == 'lgmres': 
        W_srZ = np.dot(W_sr,Z)
        ZWZ = np.dot(np.transpose(W_srZ),W_srZ)
        print(np.linalg.cond(ZWZ))
        rhs = np.dot(np.transpose(W_srZ),np.dot(W_sr,sol))
        import scipy.sparse.linalg
        s_hat, info = scipy.sparse.linalg.lgmres(ZWZ, rhs)
        for i in range(0,m1):
            for j in range(0,m2):
                surrogate[i,j] = np.dot(np.transpose(phi_tmp[i,j,:]),s_hat)

    if solver == 'lsqr': 
        W_srZ = np.dot(W_sr,Z)
        print(np.linalg.cond(W_srZ))
        rhs = np.dot(W_sr,sol)
        import scipy.sparse.linalg
        s_hat, istop, itn, r1norm, r2norm, anorm, acond, arnorm, xnorm, var = scipy.sparse.linalg.lsqr(W_srZ, rhs)
        for i in range(0,m1):
            for j in range(0,m2):
                surrogate[i,j] = np.dot(np.transpose(phi_tmp[i,j,:]),s_hat)

    if solver == 'lsmr': 
        W_srZ = np.dot(W_sr,Z)
        print(np.linalg.cond(W_srZ))
        rhs = np.dot(W_sr,sol)
        import scipy.sparse.linalg
        s_hat, lstop, itn, normr, normar, norma, conda, normx = scipy.sparse.linalg.lsmr(W_srZ, rhs,1e045)
        print(lstop)
        print(itn)
        print(normr)
        print(normar)
        print(norma)
        print(conda)
        print(normx)
        for i in range(0,m1):
            for j in range(0,m2):
                surrogate[i,j] = np.dot(np.transpose(phi_tmp[i,j,:]),s_hat)

    if solver == 'lstsq': 
        W_srZ = np.dot(W_sr,Z)
        print(np.linalg.cond(W_srZ))
        rhs = np.dot(W_sr,sol)
        import scipy.linalg
        s_hat, residues, rank, s = scipy.linalg.lstsq(W_srZ, rhs)
        print(rank)
        for i in range(0,m1):
            for j in range(0,m2):
                surrogate[i,j] = np.dot(np.transpose(phi_tmp[i,j,:]),s_hat)

    if surrogate_parameters['display']: 
        plot.surf(X, Y, surrogate, sol, fixed_value, fig, ax, current_title, True, True, 'Surrogate model','surrogate.png',d)
        #from matplotlib2tikz import save as tikz_save
        #tikz_save('surrogate.tex', figureheight = '\\figureheight', figurewidth = '\\figurewidth')
        plot.finish()

    return s_hat

def surrogate_4d(sol, Z, W_sr, s_hat, surrogate_parameters, Image, images_parameters):
    if surrogate_parameters['display']:
        import params.src.plot as plot
        fig = Image['fig']
        ax = Image['ax'] 
        current_title = Image['current_title']
        d = surrogate_parameters['d']
        M = surrogate_parameters['M']
        N = surrogate_parameters['N']

        m1 = surrogate_parameters['m1']
        m2 = surrogate_parameters['m2']

        Dom_min = surrogate_parameters['Dom_min']
        Dom_max = surrogate_parameters['Dom_max']
        P_name = surrogate_parameters['P_name']
        solver = surrogate_parameters['solver']

        normalisation = surrogate_parameters['normalisation']
        ##
        if surrogate_parameters['function_type'] == 'monomial': 
            from params.src.monomials import phi   
        elif surrogate_parameters['function_type'] == 'monomial_1norm': 
            from params.src.monomials_1norm import phi     
        ##

        shape = Z.shape
        m_max = shape[0]
        n_max = shape[1]

        x = np.linspace(-1.,1.,m1)
        y = np.linspace(-1.,1.,m2)

        surrogate = np.zeros((m1,m2))
        X0, Y0 = np.meshgrid(x, y)
        phi_tmp = np.zeros((m1,m2,n_max))

        n_c = images_parameters['number_of_cuts']

        index_c = images_parameters['index_c']
        point_c = images_parameters['point_c'] 

        c_min = sol.min()
        c_max = sol.max()

        print(index_c)
        print(point_c)

        for c in range(0,n_c):
            X = np.zeros((m1,m2))
            Y = np.zeros((m1,m2))
            Z = np.zeros((m1,m2))

            xi0 = point_c[c]
            print(index_c[c])
            for i in range(0,m1):
                for j in range(0,m2):
                    xi = np.zeros((d,))

                    if surrogate_parameters['training_points'] == 'gamma': 
                        means = surrogate_parameters['means']
                        variances = surrogate_parameters['variances']

                        X[i,j] = Dom_min[index_c[c,0]] + (Dom_max[index_c[c,0]] - Dom_min[index_c[c,0]])*(X0[i,j]+1.)/2.
                        Y[i,j] = Dom_min[index_c[c,1]] + (Dom_max[index_c[c,1]] - Dom_min[index_c[c,1]])*(Y0[i,j]+1.)/2.
                        Z[i,j] = xi0[index_c[c,2]]

                       
                        xi[index_c[c,0]] = (X[i,j]-means[index_c[c,0]]) / np.sqrt(variances[index_c[c,0]])
                        xi[index_c[c,1]] = (Y[i,j]-means[index_c[c,1]]) / np.sqrt(variances[index_c[c,1]])
                        xi[index_c[c,2]] = (Z[i,j]-means[index_c[c,2]]) / np.sqrt(variances[index_c[c,2]])

                    tmp = phi(xi,N,n_max,d)
                    surrogate[i,j] = np.dot(np.transpose(tmp),s_hat)

            if index_c[c,0] == 1:
                plot.surf4d(Z, X, Y, surrogate, c_min, c_max, fig, ax, current_title, True, True, 'Surrogate model','surrogate.png',d)
            elif index_c[c,1] == 2:
                plot.surf4d(X, Z, Y, surrogate, c_min, c_max, fig, ax, current_title, True, True, 'Surrogate model','surrogate.png',d)
            else:
                plot.surf4d(X, Y, Z, surrogate, c_min, c_max, fig, ax, current_title, True, True, 'Surrogate model','surrogate.png',d)
        plot.finish()
