/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// SWIG input file of the 'flow' module

%feature("autodoc","1");

%module(docstring=
"'floww' module: full potential solver 2018-2019
(c) ULg - A&M",
directors="1",
threads="1"
) floww
%{

#include <string>
#include <sstream>
#include <typeinfo>
#include "flow.h"

#include "fwkw.h"
#include "tboxw.h"
#include "floww.h"

%}


%include "fwkw.swg"

// ----------- MODULES UTILISES ------------
%import "tboxw.i"

// ----------- FLOW CLASSES ----------------
%include "flow.h"

%shared_ptr(flow::Medium);
%shared_ptr(flow::Assign);
%shared_ptr(flow::Initial);
%shared_ptr(flow::Dirichlet);
%shared_ptr(flow::Freestream);
%shared_ptr(flow::Slip);
%shared_ptr(flow::Wake);
%shared_ptr(flow::Kutta);
%shared_ptr(flow::Blowing);
%shared_ptr(flow::Boundary);
%shared_ptr(flow::Problem);
%shared_ptr(flow::Solver);
%shared_ptr(flow::Picard);
%shared_ptr(flow::Newton);
%shared_ptr(flow::Adjoint);

%include "wMedium.h"
%include "wAssign.h"
%include "wFreestream.h"
%warnfilter(509); // Shadowed constructor. Code compiling and running.
%include "wWake.h"
%include "wKutta.h"
%include "wBoundary.h"
%include "wBlowing.h"
%warnfilter(+509);
%include "wWakeElement.h"
%include "wKuttaElement.h"

%feature("director") flow::F0PsC;
%pythonappend flow::F0PsC::F0PsC "self.__disown__()"    // only for directors
%feature("director") flow::F0PsPhiInf;
%pythonappend flow::F0PsPhiInf::F0PsPhiInf "self.__disown__()"    // only for directors
%nodefault flow::F0Ps;     // is pure virtual but swig doesn't know it.
%include "wF0Ps.h"

%feature("director") flow::F0ElRho;
%pythonappend flow::F0ElRho::F0ElRho "self.__disown__()"    // only for directors
%feature("director") flow::F0ElRhoL;
%pythonappend flow::F0ElRhoL::F0ElRhoL "self.__disown__()"    // only for directors
%feature("director") flow::F0ElMach;
%pythonappend flow::F0ElMach::F0ElMach "self.__disown__()"    // only for directors
%feature("director") flow::F0ElMachL;
%pythonappend flow::F0ElMachL::F0ElMachL "self.__disown__()"    // only for directors
%feature("director") flow::F0ElCp;
%pythonappend flow::F0ElCp::F0ElCp "self.__disown__()"    // only for directors
%feature("director") flow::F0ElCpL;
%pythonappend flow::F0ElCpL::F0ElCpL "self.__disown__()"    // only for directors
%include "wF0El.h"

%feature("director") flow::F1ElVi;
%pythonappend flow::F1ElVi::F1ElVi "self.__disown__()"    // only for directors
%include "wF1El.h"

%immutable flow::Problem::msh; // avoids the creation of the setter method
%include "wProblem.h"

%immutable flow::Solver::pbl; // avoids the creation of the setter method
%include "wSolver.h"
%include "wPicard.h"
%include "wNewton.h"

%immutable flow::Adjoint::sol; // avoids the creation of the setter method
%include "wAdjoint.h"
