/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// global header of the "flow" module

#ifndef FLOW_H
#define FLOW_H

#if defined(WIN32)
#ifdef flow_EXPORTS
#define FLOW_API __declspec(dllexport)
#else
#define FLOW_API __declspec(dllimport)
#endif
#else
#define FLOW_API
#endif

#include "tbox.h"

/**
 * @brief Namespace for flow module
 */
namespace flow
{

// Problem and B.C. handling
class Problem;
class Medium;
class Assign;
class Initial;
class Dirichlet;
class Freestream;
class Wake;
class WakeElement;
class Kutta;
class KuttaElement;
class Blowing;
class Boundary;

// Solver
class Solver;
class Picard;
class Newton;
class FpeLSFunction;
class Adjoint;

// General
class F0Ps;
class F0El;
class F1El;
class Face;
class FaceEq;

} // namespace flow

#endif //FLOW_H
