#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Base class representing a physical boundary
# Amaury Bilocq

import numpy as np

class Boundary: 
    def __init__(self, _boundary):
        self.boundary = _boundary # boundary of interest
        self.nN = len(self.boundary.nodes) # number of nodes
        self.nE = len(self.boundary.tag.elems) # number of elements
        self.v = np.zeros((self.nN, 3)) # velocity at edge of BL
        self.u = np.zeros(self.nE) # blowing velocity
        self.Me = np.zeros(self.nN) # Mach number at the edge of the boundary layer 
        self.rhoe = np.zeros(self.nN) # density at the edge of the boundary layer
        self.connectListnodes = np.zeros(self.nN) # connectivity for nodes
        self.connectListElems = np.zeros(self.nE) # connectivity for elements
        self.deltaStar = np.zeros(self.nN) # displacement thickness
        self.xx = np.zeros(self.nN) # coordinates in the reference of the physical surface. Used to store the values for the interaction law
