/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// SWIG input file of the 'heat' module

%feature("autodoc","1");

%module(docstring=
"'heatw' module: projet MP 2015/2016
(c) ULg - A&M",
directors="1",
threads="1"
) heatw
%{

#include <string>
#include <sstream>
#include <typeinfo>
#include "heat.h"

#include "wProblem.h"

#include "wBoundary.h"
#include "wMedium.h"
#include "wSource.h"
#include "wDirichlet.h"
#include "wPeriodic.h"

#include "wSolver.h"

#include "wCompiledFct.h"

#include "wDisplayHook.h"

#include "fwkw.h"
#include "tboxw.h"

#include "wExtractor.h"

%}


%include "fwkw.swg"

// ----------- MODULES UTILISES ------------
%import "tboxw.i"

// ----------- HEAT CLASSES ----------------
%include "heat.h"

%shared_ptr(heat::Medium);
%shared_ptr(heat::Source);
%shared_ptr(heat::Dirichlet);
%shared_ptr(heat::Boundary);
%shared_ptr(heat::Periodic);
%shared_ptr(heat::Problem);
%shared_ptr(heat::Solver);



%feature("director") heat::DisplayHook;
%include "wDisplayHook.h"

%feature("director:except") {
    if ($error != NULL) {
        std::cout << "[in director:except]\n";
        //throw Swig::DirectorMethodException();
        throw std::runtime_error("Director problem");
    }
}

%include "wDirichlet.h"
%include "wMedium.h"
%include "wSource.h"
%include "wBoundary.h"
%include "wPeriodic.h"
%include "wCompiledFct.h"
%include "wExtractor.h"

%immutable heat::Problem::msh; // avoids the creation of the setter method
%include "wProblem.h"

%immutable heat::Solver::pbl; // avoids the creation of the setter method
%immutable heat::Solver::tms; // avoid the creation of the setter method
%include "wSolver.h"
