/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WMEDIUM_H
#define WMEDIUM_H

#include "heat.h"
#include "wGroup.h"
#include "wFct0.h"

using namespace tbox;

namespace heat
{

/**
 * @brief a thermal medium (with k, rhoc)
 */

class HEAT_API Medium : public Group
{
public:
#ifndef SWIG
    Fct2 &k;
    double rhoc;
#endif
    Medium(std::shared_ptr<MshData> _msh, int no, Fct2 &_k, double _rhoc = 1.0);
    Medium(std::shared_ptr<MshData> _msh, std::string const &name, Fct2 &_k, double _rhoc = 1.0);
    virtual ~Medium() { std::cout << "~Medium()\n"; }

#ifndef SWIG
    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace heat

#endif //WMEDIUM_H
