#!/bin/bash
# Submission script for Blake
#SBATCH --job-name=Mirror_all
#SBATCH --time=41:00:00 # hh:mm:ss
#
#SBATCH -N 2
#SBATCH -p blake
#
#SBATCH --comment=test
#SBATCH --output=workspace/Mirror_all.%j.out

. ~/.bash_profile $SLURM_SUBMIT_HOST

export n_nodes=2

ensemble_sizes=( 1 8 16 32)

export MKL_DYNAMIC=TRUE
export OMP_DYNAMIC=FALSE
export OMP_NUM_THREADS=24
export KMP_PLACE_THREADS=24c,1t
export KMP_AFFINITY=compact

export MKL_NUM_THREADS=${OMP_NUM_THREADS}
export N_MPI_PER_SOCKET=1

n_MPI=$(( 2*n_nodes ))
base_dir=$(pwd)
msh_file=${base_dir}/katoptron/Juelich/mesh/fused_mirror_assembly_all_mm.msh
workspace_dir=${base_dir}/workspace/katoptron_Juelich_tests_M1_assembly_fused_all
msh_part_file=${workspace_dir}/fused_mirror_assembly_all_mm_part_${n_MPI}.msh

if [ ! -d ${workspace_dir} ] 
then
    mkdir ${workspace_dir}
fi

if [ ! -e ${msh_part_file} ]
then
    gmsh -3 ${msh_file} -o ${msh_part_file} -part ${n_MPI}
fi

for s in "${ensemble_sizes[@]}"
do
    export ensemble_size=$s
    mpirun -npersocket ${N_MPI_PER_SOCKET} python run.py -k ${OMP_NUM_THREADS} katoptron/Juelich/tests/M1_assembly_fused_all.py
done
