#include "IterativeSolver.h"
#include "wTimer.h"
#include "wProblem.h"
#include "wTag.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wMedium.h"
#include "wDirichlet.h"
#include "wContact.h"
#include <map>
#include "wTag.h"
#include <math.h>
#include "wHex8.h"
#include "wQuad4.h"
#include "wTri3.h"
#include "wCacheHex8.h"
#include "wGaussHex8.h"
#include "wResults.h"

#include "wRandomVariable.h"

//#include "wElemMatrix.h"

#include "wDisplayHook.h"

#include <math.h>

#include <mpi.h>

//#include <tbb/tbb.h>
//#include <tbb/parallel_for.h>
//#include <tbb/blocked_range.h>
//#include <tbb/task_scheduler_init.h>
//#include <tbb/spin_mutex.h>

#include <cstring>
#include <sstream>
#include <iostream>
#include <locale> // std::locale

// Tpetra
#include "Stokhos_Tpetra_MP_Vector.hpp"
#include "Stokhos_Tpetra_Utilities_MP_Vector.hpp"
#include "Tpetra_ConfigDefs.hpp"

#include "Tpetra_Map.hpp"
#include "Tpetra_MultiVector.hpp"
#include "Tpetra_Vector.hpp"
#include "Tpetra_CrsGraph.hpp"
#include "Tpetra_CrsMatrix.hpp"
#include "Stokhos_Tpetra_CG.hpp"
//#include "Sacado_Fad_MP_Vector.hpp"

#include "Teuchos_XMLParameterListCoreHelpers.hpp"
#include <Teuchos_Array.hpp>
#include <Teuchos_GlobalMPISession.hpp>
#include <Teuchos_oblackholestream.hpp>
#include <Teuchos_ScalarTraits.hpp>
#include <Teuchos_RCP.hpp>
#include "Teuchos_ParameterList.hpp"

#include "Stokhos_Amesos2_MP_Vector.hpp"
#include "Amesos2_Factory.hpp"

#include "Stokhos_Ifpack2_MP_Vector.hpp"
#include "Ifpack2_Factory.hpp"

#include "Belos_Tpetra_MP_Vector.hpp"
#include "BelosStatusTestCombo.hpp"

#include <Xpetra_BlockedCrsMatrix.hpp>
#include <Xpetra_Matrix.hpp>
#include <BelosXpetraAdapter.hpp> // => This header defines Belos::XpetraOp
#include <BelosMueLuAdapter.hpp>  // => This header defines Belos::MueLuOp
#include <Xpetra_TpetraMultiVector.hpp>

#include <Xpetra_MapExtractorFactory.hpp>
#include <Xpetra_MatrixUtils.hpp>
//#include "BelosGmresSolMgr.hpp"

#include "Kokkos_ViewFactory.hpp"
#include "Kokkos_Vector.hpp"

#include <Epetra_Map.h>
#include "Epetra_MpiComm.h"

#include "BelosConfigDefs.hpp"

#include "Map.h"
#include "DiscreteProblem.h"

#include <fstream>

#include <MueLu.hpp>

#include <MueLu_Exceptions.hpp>
#include <MueLu_Hierarchy.hpp>
#include <MueLu_MasterList.hpp>
#include <MueLu_MLParameterListInterpreter.hpp>
#include <MueLu_ParameterListInterpreter.hpp>
#include <MueLu_Utilities.hpp>
#include <MueLu_HierarchyUtils.hpp>

#include <BelosXpetraStatusTestGenResSubNorm_MP_Vector.hpp>
#include <BelosStatusTestWeightedGenResNorm_MP_Vector.hpp>

#include "ElementMatrices.h"

#include "NodesList.h"

#ifdef max
#undef max // pour std::numeric_limits<std::streamsize>::max()
#endif

using namespace katoptron;
using namespace fwk;

#define HAVE_BELOS_TPETRA_TIMERS

template <typename scalar>
Teuchos::ParameterList convertParamList(Teuchos::ParameterList &paramList)
{
    for (Teuchos::ParameterList::ConstIterator i = paramList.begin(); i != paramList.end(); ++i)
    {
        const std::string &name_i = paramList.name(i);

        if (Teuchos::isParameterType<double>(paramList, name_i))
        {
            double value = Teuchos::getParameter<double>(paramList, name_i);
            std::string name_ii = name_i;

            paramList.set<scalar>(name_ii, ((scalar)value));
        }
        else if (Teuchos::isParameterType<Teuchos::ParameterList>(paramList, name_i))
        {
            Teuchos::ParameterList subparamList = Teuchos::getParameter<Teuchos::ParameterList>(paramList, name_i);
            subparamList = convertParamList<scalar>(subparamList);
            paramList.set(name_i, subparamList);
        }
    }

    return paramList;
}

/**
 * @brief Display IterativeSolver information.
 */
void IterativeSolver::write(std::ostream &out) const
{
    out << "katoptron::Solver:\n";
    out << pbl; // wProblem.h
}

/**
 * @brief Launch the solver.
 */
void IterativeSolver::start()
{
    std::locale::global(std::locale::classic());
    {
        if (minimum_ensemble_size <= 1)
        {
            IterativeSolver::tstart<double>();
        }
        else if (minimum_ensemble_size <= 8)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int, double, 8, Kokkos::OpenMP>> ensemble_scalar;
            IterativeSolver::tstart<ensemble_scalar>();
        }
        else if (minimum_ensemble_size <= 16)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int, double, 16, Kokkos::OpenMP>> ensemble_scalar;
            IterativeSolver::tstart<ensemble_scalar>();
        }
        else if (minimum_ensemble_size <= 24)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int, double, 24, Kokkos::OpenMP>> ensemble_scalar;
            IterativeSolver::tstart<ensemble_scalar>();
        }
        else if (minimum_ensemble_size <= 32)
        {
            typedef Sacado::MP::Vector<Stokhos::StaticFixedStorage<int, double, 32, Kokkos::OpenMP>> ensemble_scalar;
            IterativeSolver::tstart<ensemble_scalar>();
        }
    }
}

/**
 * @brief Launch the solver with a templated scalar type.
 */
template <typename scalar>
void IterativeSolver::tstart()
{
    LinearSolver::getTimers().reset();
    // ------------------------------------------
    //
    //         Trilinos initialization
    //
    // ------------------------------------------
    using std::cerr;
    using std::cout;
    using std::endl;
    using Teuchos::arcp;
    using Teuchos::Array;
    using Teuchos::ArrayRCP;
    using Teuchos::ArrayView;
    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;

    using Kokkos::ALL;
    using Kokkos::subview;

    typedef Map::local_ordinal_type local_ordinal_type;
    typedef Map::global_ordinal_type global_ordinal_type;
    typedef Map::node_type node_type;

    typedef typename Vectors<scalar>::tpetra_vector_type tpetra_vector_type;
    typedef typename Vectors<scalar>::tpetra_mvector_type tpetra_mvector_type;
    typedef typename Vectors<scalar>::xpetra_mvector_type xpetra_mvector_type;
    typedef typename Vectors<scalar>::xpetra_tmvector_type xpetra_tmvector_type;

    typedef typename Matrices<scalar>::tpetra_crs_type tpetra_crs_type;

    typedef typename Matrices<scalar>::xpetra_bcrs_type xpetra_bcrs_type;
    typedef typename Matrices<scalar>::xpetra_mat_type xpetra_mat_type;

    typedef Tpetra::Export<local_ordinal_type, global_ordinal_type, node_type> export_type;
    typedef Tpetra::Import<local_ordinal_type, global_ordinal_type, node_type> import_type;

    typedef MueLu::Hierarchy<scalar, local_ordinal_type, global_ordinal_type, node_type> Hierarchy;
    typedef MueLu::ParameterListInterpreter<scalar, local_ordinal_type, global_ordinal_type, node_type> ParameterListInterpreter;

    typedef Teuchos::ScalarTraits<scalar> STS;
    typedef typename STS::magnitudeType magnitude_type;

    Teuchos::oblackholestream blackhole;

    RCP<const Teuchos::Comm<int>> comm = rcp(new Teuchos::MpiComm<int>(pbl.comm));

    const size_t myRank = comm->getRank();

    // Make an output stream (for verbose output) that only prints on
    // Proc 0 of the communicator.
    Teuchos::oblackholestream blackHole;
    std::ostream &out = (myRank == 0) ? std::cout : blackHole;

    // Create some timers used by Belos so we can access them
    Teuchos::RCP<Teuchos::Time> time_mat_vec =
        Teuchos::TimeMonitor::getNewTimer("Belos: Operation Op*x");
    Teuchos::RCP<Teuchos::Time> time_prec_apply =
        Teuchos::TimeMonitor::getNewTimer("Belos: Operation Prec*x");
    Teuchos::RCP<Teuchos::Time> time_prec_setup =
        Teuchos::TimeMonitor::getNewTimer("Total MueLu setup time");

    std::string ortho_strategy = solverList->get("Orthogonalization", "DGKS");
    std::string belos_ortho_timer_name = "Belos: " + ortho_strategy;

    if (ortho_strategy == "IMGS")
        belos_ortho_timer_name = belos_ortho_timer_name + "[1]";
    else
        belos_ortho_timer_name = belos_ortho_timer_name + "[2]";

    Teuchos::RCP<Teuchos::Time> time_ortho =
        Teuchos::TimeMonitor::getNewTimer(belos_ortho_timer_name + ": Orthogonalization");
    Teuchos::RCP<Teuchos::Time> time_ortho_inner =
        Teuchos::TimeMonitor::getNewTimer(belos_ortho_timer_name + ": Ortho (Inner Product)");
    Teuchos::RCP<Teuchos::Time> time_ortho_norm =
        Teuchos::TimeMonitor::getNewTimer(belos_ortho_timer_name + ": Ortho (Norm)");
    Teuchos::RCP<Teuchos::Time> time_ortho_update =
        Teuchos::TimeMonitor::getNewTimer(belos_ortho_timer_name + ": Ortho (Update)");

    Teuchos::RCP<Teuchos::Time> time_belos_total =
        Teuchos::TimeMonitor::getNewTimer("Belos: PseudoBlockGmresSolMgr total solve time");

    bool write_intermediate_files = solverList->get("Write intermediate files", false);
    bool write_txt_files = solverList->get("Write txt files", false);
    bool write_vtk_files = solverList->get("Write vtk files", true);
    {
        LinearSolver::getTimers()["total"].start();
        auto msh = pbl.msh;

        // assign a global row for each global node
        for (auto n : msh->nodes)
            n->row = n->no - 1;

        auto numNodes = msh->nodes.size();

        RCP<Teuchos::ParameterList> randomParams =
            rcp(new Teuchos::ParameterList(solverList->get("randomParams", Teuchos::ParameterList())));
        const int n_rv = pbl.RandomVariables.size();

        typedef Kokkos::View<scalar *, Kokkos::LayoutLeft> RandomVariableView;

        RandomVariableView m_rv_values = RandomVariableView("KL Random Variables", n_rv);

        typedef EnsembleTraits<scalar> ET;
        const int ensemble_size = ET::size;

        for (int i = 0; i < n_rv; ++i)
            for (int j = 0; j < ensemble_size; ++j)
                ET::coeff(m_rv_values(i), j) = (double)pbl.RandomVariables[i]->values[j];

        LinearSolver::getTimers()["dp"].start();
        Teuchos::RCP<DiscreteProblem<scalar>> discreteProblem =
            Teuchos::rcp(new DiscreteProblem<scalar>(pbl, numPrimalDPN, randomParams, m_rv_values));
        LinearSolver::getTimers()["dp"].stop();
        LinearSolver::getTimers()["computeMatrices"].start();

        const bool contactProblem = (discreteProblem->domain->contactsList->getContactNumber() >= 1);

        if (contactProblem)
        {
            LinearSolver::getTimers()["compute contact matrices"].start();
            discreteProblem->computeContactMatrices(pbl.comm);
            LinearSolver::getTimers()["compute contact matrices"].stop();
        }
        LinearSolver::getTimers()["compute block 00 matrix"].start();
        discreteProblem->computeMatrices();
        LinearSolver::getTimers()["compute block 00 matrix"].stop();
        LinearSolver::getTimers()["computeMatrices"].stop();

        LinearSolver::getTimers()["computeMatrices"].start();

        const bool mergeMatrix = solverList->get("Merge block matrix", true);
        const bool writeMatrixAndVectors = solverList->get("Write matrix and vectors", false);

        if (writeMatrixAndVectors)
        {
            LinearSolver::getTimers()["write"].start();
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeSparseFile(std::string("A_mm.txt"),
                                                                           discreteProblem->algebraic->matrices->K);
            if (contactProblem)
            {
                Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeSparseFile(std::string("B_mm.txt"),
                                                                               discreteProblem->algebraic->matrices->B);
                Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeSparseFile(std::string("B_T_mm.txt"),
                                                                               discreteProblem->algebraic->matrices->B_T);
            }
            LinearSolver::getTimers()["write"].stop();
        }

        LinearSolver::getTimers()["computeLoads"].start();
        bool is_scaled = solverList->get("Scaled Block 0", false);
        discreteProblem->computeLoads(is_scaled);
        LinearSolver::getTimers()["computeLoads"].stop();

        if (contactProblem)
        {
            scalar gamma;
            if (solverList->isType<std::string>("gamma"))
            {
                Teuchos::Array<double> array_gamma_double = Teuchos::getArrayFromStringParameter<double>(*solverList, "gamma");
                for (int j = 0; j < ensemble_size; ++j)
                    ET::coeff(gamma, j) = (double)array_gamma_double[j];
            }
            else
                gamma = solverList->get("gamma", 0.);
            discreteProblem->createBlockMatrix(gamma, writeMatrixAndVectors, mergeMatrix);
        }

        LinearSolver::getTimers()["computeMatrices"].stop();

        LinearSolver::getTimers()["computeLoads"].start();
        if (contactProblem)
        {
            discreteProblem->createBlockMVector(writeMatrixAndVectors);
        }
        LinearSolver::getTimers()["computeLoads"].stop();

        if (writeMatrixAndVectors)
        {
            std::string name_1, name_2, name_3;
            name_1 = "matrix_after_bc.text";
            name_2 = "matrix_after_bc";
            name_3 = "vector_after_bc.text";

            LinearSolver::getTimers()["write"].start();
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeSparseFile(std::string("A_mm_after_bc.txt"),
                                                                           discreteProblem->algebraic->matrices->K);
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("b_mm_after_bc.txt"),
                                                                          discreteProblem->algebraic->vectors->b);
            LinearSolver::getTimers()["write"].stop();
        }

        std::string name_4, defaultName, name_x;
        defaultName = "none";
        name_4 = solverList->get("Initial guess", defaultName);
        name_x = "solution";

        RCP<export_type> exportOutput;
        RCP<import_type> importOutput;
        RCP<tpetra_vector_type> xPost, xPost2;
        if (contactProblem)
        {
            importOutput = rcp(new import_type(discreteProblem->algebraic->map->fullmap, discreteProblem->algebraic->map->fullmapWO));
            exportOutput = rcp(new export_type(discreteProblem->algebraic->map->fullmapWO, discreteProblem->algebraic->map->fullmapOutput));
            xPost = rcp(new tpetra_vector_type(discreteProblem->algebraic->map->fullmapWO, false));
            xPost2 = rcp(new tpetra_vector_type(discreteProblem->algebraic->map->fullmapOutput, false));
        }
        else
        {
            importOutput = rcp(new import_type(discreteProblem->algebraic->map->mapDofs, discreteProblem->algebraic->map->mapDofsWO));
            exportOutput = rcp(new export_type(discreteProblem->algebraic->map->mapDofsWO, discreteProblem->algebraic->map->mapOutput));
            xPost = rcp(new tpetra_vector_type(discreteProblem->algebraic->map->mapDofsWO, false));
            xPost2 = rcp(new tpetra_vector_type(discreteProblem->algebraic->map->mapOutput, false));
        }

        if (name_4.compare(defaultName) != 0)
            discreteProblem->algebraic->vectors->x = discreteProblem->algebraic->readVectorFromFile(name_4, myRank);

        RCP<tpetra_vector_type> r(new tpetra_vector_type(discreteProblem->algebraic->map->mapDofs, true));

        discreteProblem->algebraic->matrices->K->apply(*(discreteProblem->algebraic->vectors->x), *r, Teuchos::NO_TRANS, (scalar)-1., (scalar)1.);

        if (writeMatrixAndVectors)
        {
            LinearSolver::getTimers()["write"].start();
            std::string name_5 = "x_0.text";
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string(name_5), discreteProblem->algebraic->vectors->x);
            std::string name_6 = "r_0.text";
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string(name_6), r);
            LinearSolver::getTimers()["write"].stop();
        }

        bool usePrec = solverList->get("Use preconditioner", contactProblem);

        std::string belos_output_file = solverList->get("Output Stream file name belos", "belos_out.txt");

        std::ofstream belos_ofstream(belos_output_file);

        RCP<Teuchos::FancyOStream> verbOut = Teuchos::getFancyOStream(Teuchos::rcpFromRef(belos_ofstream));
        solverList->set("Output Stream", verbOut->getOStream());

        magnitude_type GMRES_tol = solverList->get("Ensemble Convergence Tolerance", 1E-08);
        solverList->set("Convergence Tolerance", GMRES_tol);

        active_set_iteration = 0;
        size_t max_active_set_iteration = solverList->get("Maximum active set iteration", 1);

        bool blockedStatusTest = solverList->get("Use blocked status test", !mergeMatrix);
        bool blockedStatusTest_x = solverList->get("Use blocked status test: x", blockedStatusTest);
        bool blockedStatusTest_T = solverList->get("Use blocked status test: T", blockedStatusTest);

        LinearSolver::getTimers()["solver"].start();
        while (true)
        {
            // Solve
            RCP<tpetra_vector_type> xx;

            const bool blockedMatrix = (numPrimalDPN == 4);

            RCP<xpetra_mat_type> A;
            if (contactProblem)
                A = discreteProblem->algebraic->matrices->blockedMatrix;
            else if (mergeMatrix && blockedMatrix)
                A = Teuchos::rcp_dynamic_cast<xpetra_bcrs_type>(discreteProblem->algebraic->matrices->A)->Merge();
            else
                A = discreteProblem->algebraic->matrices->A;

            typedef xpetra_mvector_type MV;
            typedef typename tpetra_mvector_type::dot_type belos_scalar;
            typedef Belos::OperatorT<MV> OP;

            typedef Belos::LinearProblem<belos_scalar, MV, OP> BLinProb;
            // Turns a Xpetra::Matrix object into a Belos operator
            RCP<OP> belosOp = rcp(new Belos::XpetraOp<scalar, local_ordinal_type, global_ordinal_type, node_type>(A));
            RCP<OP> belosPrec;

            if (usePrec)
            {
                LinearSolver::getTimers()["preconditioner setup"].start();
                RCP<xpetra_mvector_type> nullspace1 = Teuchos::null;
                RCP<xpetra_mvector_type> nullspace2 = Teuchos::null;

                bool useXML = solverList->get("use xml file", false);
                string xmlFileName = solverList->get("MueLu xml file name", " ");
                RCP<Teuchos::ParameterList> mueluParams;

                if (useXML || contactProblem)
                {
                    bool doIconver = solverList->get("convert MueLu xml file", false);

                    RCP<Teuchos::ParameterList> paramList = Teuchos::getParametersFromXmlFile(xmlFileName);

                    Teuchos::ParameterList paramList2;

                    if (contactProblem)
                    {
                        RCP<std::map<local_ordinal_type, local_ordinal_type>> lagr2dof =
                            Teuchos::rcpFromRef(discreteProblem->algebraic->map->lm_to_dof);

                        RCP<Teuchos::ParameterList> paramsF, paramsI;
                        paramsF = sublist(paramList, "Factories");
                        paramsI = sublist(paramsF, "myInterfaceAggs2");
                        paramsI->set<RCP<std::map<local_ordinal_type, local_ordinal_type>>>("DualNodeID2PrimalNodeID - level 0", lagr2dof);
                    }

                    if (doIconver)
                        paramList2 = convertParamList<scalar>(*paramList);
                    else
                        paramList2 = (*paramList);

                    mueluParams = rcp(new Teuchos::ParameterList(paramList2));
                }
                else
                    mueluParams = rcp(new Teuchos::ParameterList(solverList->get("mueluParams", Teuchos::ParameterList())));

                ParameterListInterpreter mueLuFactory(*mueluParams, comm);
                RCP<Hierarchy> H = mueLuFactory.CreateHierarchy();
                H->GetLevel(0)->Set("A", A);

                H->IsPreconditioner(true);
                H->SetDefaultVerbLevel(MueLu::Extreme);
                mueLuFactory.SetupHierarchy(*H);
                LinearSolver::getTimers()["preconditioner setup"].stop();

                // Turns a MueLu::Hierarchy object into a Belos operator
                belosPrec = rcp(new Belos::MueLuOp<scalar, local_ordinal_type, global_ordinal_type, node_type>(H));
            }

            RCP<MV> xMV;
            RCP<MV> bMV;
            RCP<MV> weightsMV;

            if (contactProblem)
            {
                xMV = discreteProblem->algebraic->vectors->solutionMultiVector;
                bMV = discreteProblem->algebraic->vectors->rhsMultiVector;
            }
            else
            {
                xMV = rcp(new xpetra_tmvector_type(discreteProblem->algebraic->vectors->x));
                bMV = rcp(new xpetra_tmvector_type(discreteProblem->algebraic->vectors->b));
                weightsMV = rcp(new xpetra_tmvector_type(discreteProblem->algebraic->vectors->weights));
            }

            RCP<BLinProb> blinproblem = rcp(new BLinProb(belosOp, xMV, bMV));

            if (usePrec)
                blinproblem->setRightPrec(belosPrec);
            blinproblem->setProblem();
            RCP<Teuchos::ParameterList> belosParams = solverList;
            RCP<Belos::SolverManager<belos_scalar, MV, OP>> blinsolver =
                rcp(new Belos::PseudoBlockGmresSolMgr<belos_scalar, MV, OP>(blinproblem, belosParams));
            out << "Belos informations are written in file " << belos_output_file << endl;

            if (writeMatrixAndVectors)
            {
                LinearSolver::getTimers()["write"].start();
                if (contactProblem)
                {
                    RCP<tpetra_mvector_type> mxx =
                        (Teuchos::rcp_dynamic_cast<xpetra_tmvector_type>(discreteProblem->algebraic->vectors->solutionMultiVector))->getTpetra_MultiVector();
                    RCP<tpetra_mvector_type> mbb =
                        (Teuchos::rcp_dynamic_cast<xpetra_tmvector_type>(discreteProblem->algebraic->vectors->rhsMultiVector))->getTpetra_MultiVector();
                    Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("blocked_x_0_mm.txt"), mxx->getVectorNonConst(0));
                    Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("blocked_b_mm.txt"), mbb->getVectorNonConst(0));
                }
                LinearSolver::getTimers()["write"].stop();
            }

            if (contactProblem && solverList->get("Create Preconditioned Matrix", false))
            {
                LinearSolver::getTimers()["create AM^{-1}"].start();
                typedef xpetra_tmvector_type TMV;

                Teuchos::RCP<const Xpetra::Map<local_ordinal_type, global_ordinal_type, node_type>> merged_map =
                    discreteProblem->algebraic->vectors->rhsMultiVector->getMap();

                Teuchos::RCP<TMV> MVx1 = Teuchos::rcp(new TMV(merged_map, merged_map->getGlobalNumElements(), true));
                Teuchos::RCP<TMV> MVx2 = Teuchos::rcp(new TMV(merged_map, merged_map->getGlobalNumElements(), true));
                Teuchos::RCP<TMV> MVy1 = Teuchos::rcp(new TMV(merged_map, merged_map->getGlobalNumElements(), true));
                Teuchos::RCP<TMV> MVy2 = Teuchos::rcp(new TMV(merged_map, merged_map->getGlobalNumElements(), true));

                for (size_t local_i = 0; local_i < merged_map->getNodeNumElements(); ++local_i)
                    MVx1->sumIntoGlobalValue(merged_map->getGlobalElement(local_i), merged_map->getGlobalElement(local_i), ((scalar)1.));

                belosPrec->Apply(*MVx1, *MVx2);
                belosOp->Apply(*MVx2, *MVy2);
                belosOp->Apply(*MVx1, *MVy1);

                Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("Applied_Matrix_mm.txt"),
                                                                              MVy1->getTpetra_MultiVector());
                Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("Applied_Preconditioner_mm.txt"),
                                                                              MVx2->getTpetra_MultiVector());
                Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("Applied_Preconditioned_Matrix_mm.txt"),
                                                                              MVy2->getTpetra_MultiVector());

                LinearSolver::getTimers()["create AM^{-1}"].stop();
            }

            if (blockedMatrix && blockedStatusTest)
            {
                typedef Belos::StatusTestGenResSubNorm<belos_scalar, MV, OP> blockStatusTestClass;
                typedef Belos::StatusTestWeightedGenResNorm<belos_scalar, MV, OP> weightedStatusTestClass;
                typedef Belos::StatusTestCombo<belos_scalar, MV, OP> StatusTestComboClass;

                magnitude_type GMRES_tol_x = solverList->get("Ensemble Convergence Tolerance: x", 1E-08);
                magnitude_type GMRES_tol_T = solverList->get("Ensemble Convergence Tolerance: T", 1E-08);
                magnitude_type GMRES_tol_w = solverList->get("Ensemble Convergence Tolerance: weights", 1E-08);

                bool x_rel = solverList->get("Ensemble Convergence Tolerance: x - relative", true);
                bool x_rel_full_rhs = solverList->get("Ensemble Convergence Tolerance: x - relative full rhs", false);
                bool T_rel = solverList->get("Ensemble Convergence Tolerance: T - relative", true);
                bool T_rel_full_rhs = solverList->get("Ensemble Convergence Tolerance: T - relative full rhs", false);

                bool weightStatusTest = solverList->get("Use weighted status test", false);
                bool w_rel = solverList->get("Ensemble Convergence Tolerance: weights - relative", true);

                RCP<blockStatusTestClass> blockStatusTest_T = Teuchos::rcp(new blockStatusTestClass(GMRES_tol_T, 0));
                RCP<blockStatusTestClass> blockStatusTest_x = Teuchos::rcp(new blockStatusTestClass(GMRES_tol_x, 1));
                RCP<weightedStatusTestClass> weightedStatusTest = Teuchos::rcp(new weightedStatusTestClass(GMRES_tol_w, weightsMV));

                if (x_rel && x_rel_full_rhs)
                    blockStatusTest_x->defineScaleForm(Belos::NormOfFullInitRes, Belos::TwoNorm);
                else if (!x_rel)
                    blockStatusTest_x->defineScaleForm(Belos::None, Belos::TwoNorm);
                if (T_rel && T_rel_full_rhs)
                    blockStatusTest_T->defineScaleForm(Belos::NormOfFullInitRes, Belos::TwoNorm);
                else if (!T_rel)
                    blockStatusTest_T->defineScaleForm(Belos::None, Belos::TwoNorm);
                if (!w_rel)
                    weightedStatusTest->defineScaleForm(Belos::None, Belos::TwoNorm);

                RCP<StatusTestComboClass> statusTestCombo = Teuchos::rcp(new StatusTestComboClass(StatusTestComboClass::SEQ));

                if (blockedStatusTest_T)
                    statusTestCombo->addStatusTest(blockStatusTest_T);
                if (blockedStatusTest_x)
                    statusTestCombo->addStatusTest(blockStatusTest_x);
                if (weightStatusTest)
                    statusTestCombo->addStatusTest(weightedStatusTest);

                blinsolver->setUserConvStatusTest(statusTestCombo);
            }
            out << "--------- Belos start" << endl;
            blinsolver->solve();
            out << "--------- Belos stop" << endl;

            if (contactProblem)
            {
                RCP<tpetra_mvector_type> mxx =
                    (Teuchos::rcp_dynamic_cast<xpetra_tmvector_type>(discreteProblem->algebraic->vectors->solutionMultiVector))->getTpetra_MultiVector();
                xx = mxx->getVectorNonConst(0);

                if (write_intermediate_files)
                {
                    LinearSolver::getTimers()["solver"].stop();
                    LinearSolver::getTimers()["write results"].start();

                    std::string name_5 = "x_sol" + std::to_string(active_set_iteration) + ".txt";
                    Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string(name_5), xx);
                    LinearSolver::getTimers()["write results"].stop();
                    LinearSolver::getTimers()["solver"].start();
                }

                xPost->doImport(*xx, *importOutput, Tpetra::INSERT);
            }
            else
            {
                xPost->doImport(*(discreteProblem->algebraic->vectors->x), *importOutput, Tpetra::INSERT);
            }
            ++active_set_iteration;

            if (write_intermediate_files)
            {
                LinearSolver::getTimers()["solver"].stop();
                LinearSolver::getTimers()["write results"].start();
                if (write_txt_files)
                {
                    xPost2->doExport(*(xPost), *exportOutput, Tpetra::INSERT);
                    Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("x_mm.txt"), xPost2);
                }
                if (write_vtk_files)
                {
                    this->writeResults<scalar>(xPost, numNodes, name_x, active_set_iteration,
                                               false, false, discreteProblem->algebraic->map,
                                               discreteProblem->domain->random_field);
                }
                LinearSolver::getTimers()["write results"].stop();
                LinearSolver::getTimers()["solver"].start();
            }

            Mask<scalar> hasConverged;
            const size_t ensemble_size = hasConverged.getSize();

            for (size_t l = 0; l < ensemble_size; ++l)
                hasConverged.set(l, true);

            if (active_set_iteration == max_active_set_iteration)
            {
                if (contactProblem)
                {
                    LinearSolver::getTimers()["update"].start();
                    hasConverged = discreteProblem->updateActivity(active_set_iteration);
                    LinearSolver::getTimers()["update"].stop();

                    if (MaskLogic::AND(hasConverged) == false)
                    {
                        out << "---------------------------------------------------------" << std::endl;
                        out << "---Has not converged before reaching maximum iteration---" << std::endl;
                        out << "---------------------------------------------------------" << std::endl;
                    }
                }
                break;
            }
            else if (contactProblem)
            {
                LinearSolver::getTimers()["update"].start();
                hasConverged = discreteProblem->updateActivity(active_set_iteration);
                LinearSolver::getTimers()["update"].stop();
            }
            if (MaskLogic::AND(hasConverged))
                break;
            else
            {
#ifndef HAVE_STOKHOS_ENSEMBLE_REDUCT
                mask_assign(hasConverged, GMRES_tol) = {1., GMRES_tol};
                solverList->set("Convergence Tolerance", GMRES_tol);
#else
                ;
#endif
            }
        }

        LinearSolver::getTimers()["solver"].stop();

        LinearSolver::getTimers()["write results"].start();
        if (write_txt_files)
        {
            xPost2->doExport(*(xPost), *exportOutput, Tpetra::INSERT);
            Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("x_mm.txt"), xPost2);
        }
        if (write_vtk_files)
        {
            this->writeResults<scalar>(xPost, numNodes, name_x, active_set_iteration, true, false,
                                       discreteProblem->algebraic->map, discreteProblem->domain->random_field);
        }
        LinearSolver::getTimers()["write results"].stop();
        // ------------------------------------------
        //
        //        End of the timers and print
        //
        // ------------------------------------------
        LinearSolver::getTimers()["total"].stop();

        discreteProblem->domain->elementsList = Teuchos::null;
        discreteProblem->domain->nodesList = Teuchos::null;
        discreteProblem->domain->materialsList = Teuchos::null;
        discreteProblem->domain->contactsList = Teuchos::null;
        discreteProblem->domain->random_field = Teuchos::null;
        discreteProblem->domain->~Domain();
        discreteProblem = Teuchos::null;

        double time_mat_vec_sec = time_mat_vec->totalElapsedTime();
        double time_prec_setup_sec = time_prec_setup->totalElapsedTime();
        double time_prec_apply_sec = time_prec_apply->totalElapsedTime();
        double time_ortho_sec = time_ortho->totalElapsedTime();

        double time_belos_total_sec = time_belos_total->totalElapsedTime();

        double time_ortho_inner_sec = time_ortho_inner->totalElapsedTime();
        double time_ortho_norm_sec = time_ortho_norm->totalElapsedTime();
        double time_ortho_update_sec = time_ortho_update->totalElapsedTime();

        static clock_t freq = CLOCKS_PER_SEC;

        UnixClockTime zero_time = UnixClockTime(0);
        UnixClockTime time_mat_vec_time = UnixClockTime(time_mat_vec_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_prec_setup_time = UnixClockTime(time_prec_setup_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_prec_apply_time = UnixClockTime(time_prec_apply_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_ortho_time = UnixClockTime(time_ortho_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_ortho_inner_time = UnixClockTime(time_ortho_inner_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_ortho_norm_time = UnixClockTime(time_ortho_norm_sec * static_cast<double>(freq) / 10000);
        UnixClockTime time_ortho_update_time = UnixClockTime(time_ortho_update_sec * static_cast<double>(freq) / 10000);

        UnixClockTime time_belos_total_time = UnixClockTime(time_belos_total_sec * static_cast<double>(freq) / 10000);

        Times time_mat_vec_times, time_prec_setup_times, time_prec_apply_times, time_ortho_times,
            time_ortho_inner_times, time_ortho_norm_times, time_ortho_update_times;
        Times zero_times;

        Times time_belos_total_times;

        time_mat_vec_times.setUser(zero_time);
        time_mat_vec_times.setReal(time_mat_vec_time);
        time_mat_vec_times.setKernel(zero_time);

        time_prec_setup_times.setUser(zero_time);
        time_prec_setup_times.setReal(time_prec_setup_time);
        time_prec_setup_times.setKernel(zero_time);

        time_prec_apply_times.setUser(zero_time);
        time_prec_apply_times.setReal(time_prec_apply_time);
        time_prec_apply_times.setKernel(zero_time);

        time_ortho_times.setUser(zero_time);
        time_ortho_times.setReal(time_ortho_time);
        time_ortho_times.setKernel(zero_time);

        time_ortho_inner_times.setUser(zero_time);
        time_ortho_inner_times.setReal(time_ortho_inner_time);
        time_ortho_inner_times.setKernel(zero_time);

        time_ortho_norm_times.setUser(zero_time);
        time_ortho_norm_times.setReal(time_ortho_norm_time);
        time_ortho_norm_times.setKernel(zero_time);

        time_ortho_update_times.setUser(zero_time);
        time_ortho_update_times.setReal(time_ortho_update_time);
        time_ortho_update_times.setKernel(zero_time);

        zero_times.setUser(zero_time);
        zero_times.setReal(zero_time);
        zero_times.setKernel(zero_time);

        time_belos_total_times.setUser(zero_time);
        time_belos_total_times.setReal(time_belos_total_time);
        time_belos_total_times.setKernel(zero_time);

        LinearSolver::getTimers()["time_belos_total"].setTStart(zero_times);
        LinearSolver::getTimers()["time_belos_total"].setTElaps(time_belos_total_times);

        LinearSolver::getTimers()["time_mat_vec"].setTStart(zero_times);
        LinearSolver::getTimers()["time_mat_vec"].setTElaps(time_mat_vec_times);

        LinearSolver::getTimers()["time_prec_setup"].setTStart(zero_times);
        LinearSolver::getTimers()["time_prec_setup"].setTElaps(time_prec_setup_times);

        LinearSolver::getTimers()["time_prec_apply"].setTStart(zero_times);
        LinearSolver::getTimers()["time_prec_apply"].setTElaps(time_prec_apply_times);

        LinearSolver::getTimers()["time_ortho"].setTStart(zero_times);
        LinearSolver::getTimers()["time_ortho"].setTElaps(time_ortho_times);

        LinearSolver::getTimers()["time_ortho_inner"].setTStart(zero_times);
        LinearSolver::getTimers()["time_ortho_inner"].setTElaps(time_ortho_inner_times);

        LinearSolver::getTimers()["time_ortho_norm"].setTStart(zero_times);
        LinearSolver::getTimers()["time_ortho_norm"].setTElaps(time_ortho_norm_times);

        LinearSolver::getTimers()["time_ortho_update"].setTStart(zero_times);
        LinearSolver::getTimers()["time_ortho_update"].setTElaps(time_ortho_update_times);

        out << "\n---CPU statistics of proc " << myRank << "---\n";
        out << LinearSolver::getTimers();
        out << "---------------------------\n";

        bool print_teuchos_timers = solverList->get("Print Teuchos timers", false);
        if (print_teuchos_timers)
        {
            std::streambuf *psbuf, *backup;
            std::ofstream filestr;
            filestr.open("teuchos_timers.txt");

            backup = std::cout.rdbuf(); // back up cout's streambuf

            psbuf = filestr.rdbuf(); // get file's streambuf
            std::cout.rdbuf(psbuf);  // assign streambuf to cout

            Teuchos::TimeMonitor::summarize();

            std::cout.rdbuf(backup); // restore cout's original streambuf

            filestr.close();
        }

        Teuchos::TimeMonitor::zeroOutTimers();
    }
}
