#!/usr/bin/env python3
# -*- coding: utf-8; -*-




import katoptron as m
import tbox
import tbox.gmsh as gmsh
import katoptron.utilities as u
#import matplotlib.pyplot as plt


def main():
    comm, rank, size = u.import_MPI()

    from PyTrilinos import Teuchos

    import shlex
    import subprocess
    import os
    import numpy as np

    geo_name = '2_cubes_tet.geo'
    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    msh, mesh_name, partitioned_mesh_name = u.read_mesh_MPI(
        geo_name, file_dir, work_dir, comm, rank, size)

    ensemble_size = 1

    pbl = m.Problem(msh, comm)

    #E = np.linspace(0.5,1.5,ensemble_size)
    E = 1. * np.ones(ensemble_size)
    nu = 0.2 * np.ones(ensemble_size)
    k = 0.114
    d = 0.
    m.Medium(pbl, "Body 1", "test", E, nu, k, d)
    m.Medium(pbl, "Body 2", "test", E, nu, k, d)

    m.Dirichlet(pbl, "Surf 1 1", "Clamped", 1, 0.,
                1, 0., 1, 0., 1, 10., ensemble_size)
    m.Dirichlet(pbl, "Surf 2 2", "Clamped", 1, 0.,
                1, 0., 1, 2., 1, 20., ensemble_size)

    norm = tbox.Vector3d(0, 0, -1)
    cont = m.Contact(pbl, "Surf 1 2", "contact", norm)
    # cont.setInitialyOpen()
    cont.setNoUpdate()
    cont.setSticking()
    cont.setMaster(pbl, "Surf 2 1", norm)

    dx = np.zeros(ensemble_size)
    dy = 2. * np.ones(ensemble_size)
    dz = np.zeros(ensemble_size)
    dT = 10. * np.ones(ensemble_size)

    solverList = Teuchos.ParameterList()
    solverList['Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 1000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    solverList["MueLu xml file name"] = file_dir + '/multigrid_2_lvls_3_1.xml'
    solverList["convert MueLu xml file"] = True

    solverList['Use preconditioner'] = True

    solverList['type'] = "BlockGmres"
    solverList['Maximum active set iteration'] = 1

    solverList['Print Teuchos timers'] = True

    slv = m.IterativeSolver(pbl, solverList, 3, ensemble_size)

    nThreads = u.Initialize_Kokkos()

    slv.start()

    u.Finalize_Kokkos()


if __name__ == "__main__":
    main()
