#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



import matplotlib as mpl
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import LinearLocator, FormatStrFormatter
from numpy.polynomial import Polynomial, Legendre
def main():
    m = 100
    n = 10

    x_min = 100
    x_max = 300

    y_min = 2
    y_max = 8

    x = np.linspace(x_min,x_max,m)
    y = np.linspace(y_min,y_max,m)

    Z = np.zeros((m,m))

    X, Y = np.meshgrid(x, y)

    fig = plt.figure()
    ax = fig.gca(projection='3d')
    ax.set_xlabel('T [C]')
    ax.set_ylabel('H_1 [mm]')
    ax.view_init(elev=20., azim=-140.)
    current_title = plt.title(" ")

    index = 1

    for i in range(0,n):
        for j in range(0,n):
            z_x = Legendre.basis(i, [x_min,x_max])(x)
            z_y = Legendre.basis(j, [y_min,y_max])(y)
            for ii in range(0,m):
                for jj in range(0,m):
                    Z[ii,jj] = z_x[ii] * z_y[jj]

            surf = ax.plot_surface(X, Y, Z, rstride=1, cstride=1, cmap=cm.jet, alpha=0.4,
                       linewidth=0, antialiased=False)

            current_title.set_text('i = ' + str(i) + ', j = ' + str(j))
            plt.savefig('Leg_' + str(index) + '.png', bbox_inches = 'tight')
            index = index + 1
            surf.remove()
            del surf

    surf = ax.plot_surface(X, Y, Z, rstride=1, cstride=1, cmap=cm.jet, alpha=0.4,
               linewidth=0, antialiased=False)
if __name__ == "__main__":
    main()
