/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wGmres.h"
using namespace tbox;

Gmres::Gmres() : LinearSolver()
{
    // set the preconditioner and solver parameters to "good" starting values
    solver.preconditioner().setFillfactor(1);
    solver.preconditioner().setDroptol(1e-6);
    solver.set_restart(50);
    solver.setTolerance(1e-8);
    solver.setMaxIterations(1000);
}

void Gmres::analyze(Eigen::SparseMatrix<double> const &A)
{
    solver.analyzePattern(A);
}
void Gmres::factorize(Eigen::SparseMatrix<double> const &A)
{
    solver.factorize(A);
}
void Gmres::solve(Eigen::Map<Eigen::VectorXd> const &b, Eigen::Map<Eigen::VectorXd> &x)
{
    x = solver.solve(b);
}
void Gmres::compute(Eigen::SparseMatrix<double> const &A, Eigen::Map<Eigen::VectorXd> const &b, Eigen::Map<Eigen::VectorXd> &x)
{
    solver.compute(A);
    x = solver.solve(b);
}

void Gmres::write(std::ostream &out) const
{
    out << "GMRES (Eigen)" << std::endl;
}
