/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WHEX8_H
#define WHEX8_H

#include "tbox.h"
#include "wLazy.h"
#include "wElement.h"
#include "wCacheHex8.h"

namespace tbox
{

/**
 * @brief Linear hexaedron finite element
 * @authors Kim Liegeois, Romain Boman, Adrien Crovato
 */
class TBOX_API Hex8 : public Element
{
#ifndef SWIG
    friend class MemHex8;
    MemHex8 *pmem; ///< private memory of precalculated values
    inline static CacheHex8 &getCache();
    inline MemHex8 &getMem();
    virtual double buildJ(size_t k, Eigen::MatrixXd &J, Eigen::MatrixXd &iJ) const override;

    virtual Eigen::MatrixXd buildK(std::vector<double> const &u, Fct0 const &fct) override;
    virtual Eigen::MatrixXd buildM() override;
    virtual Eigen::MatrixXd buildK_mechanical(Eigen::MatrixXd const &H) override;
    virtual Eigen::MatrixXd buildS_thermomechanical(Eigen::MatrixXd const &D) override;
    virtual Eigen::MatrixXd buildL_thermal(Eigen::MatrixXd const &C) override;

    virtual void strain_stress(Eigen::MatrixXd &Strain, Eigen::MatrixXd &Stress, Eigen::MatrixXd const &H, std::vector<double> const &u) override;

    virtual double computeV(std::vector<double> const &u, Fct0 const &fct) override;
    virtual double computeV() override;
#endif

public:
    Hex8(int n, Tag *_ptag, Tag *_etag, std::vector<Tag *> &_parts, std::vector<Node *> &nods);
    virtual ~Hex8();
    virtual ELTYPE type() const override { return ELTYPE::HEX8; }

#ifndef SWIG
    virtual Cache &getVCache() override;
    virtual Mem &getVMem() override;
#endif
};

#include "wHex8.inl"

} // namespace tbox

#endif //WHEX8_H
