/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @brief Return the determinant of the Jacobian at Gauss point k
 */
inline double Mem::getDetJ(size_t k)
{
    if (valid)
        return detJs[k];
    else
        throw std::runtime_error("Mem::getDetJ is invalid because element memory is not up-to-date!\n");
}

/**
 * @brief Return the the Jacobian matrix at Gauss point k
 */
inline Eigen::MatrixXd const &Mem::getJ(size_t k)
{
    if (valid)
        return Js[k];
    else
        throw std::runtime_error("Mem::getJ is invalid because element memory is not up-to-date!\n");
}

/**
 * @brief Return the inverse of the Jacobian matrix at Gauss point k
 */
inline Eigen::MatrixXd const &Mem::getJinv(size_t k)
{
    if (valid)
        return Jinvs[k];
    else
        throw std::runtime_error("Mem::getJinv is invalid because element memory is not up-to-date!\n");
}

/**
 * @brief Return the volume of the element
 */
inline double Mem::getVol()
{
    if (valid)
        return vol;
    else
        throw std::runtime_error("Mem::getVol is invalid because element memory is not up-to-date!\n");
}

/**
 * @brief Return the center of gravity of the element
 */
inline Eigen::Vector3d const &Mem::getCg()
{
    if (valid)
        return cg;
    else
        throw std::runtime_error("Mem::getCg is invalid because element memory is not up-to-date!\n");
}
