/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wMemTetra4.h"
#include "wNode.h"

using namespace tbox;

MemTetra4::MemTetra4(Tetra4 &_e, size_t _ngp) : Mem(_ngp), e(_e)
{
}

/**
 * @brief Update Tetra4 values
 */
void MemTetra4::update(bool sameDim)
{
    // Jacobian
    if (!sameDim)
        throw std::runtime_error("MemTetra4::update(): Tetra4 must always have the same dimension as the problem!\n");
    // Tetra4 is always of same dimension as Problem (volume element)
    for (size_t k = 0; k < ngp; ++k)
    {
        detJs[k] = e.buildJ(k, Js[k], Jinvs[k]);
        checkJac(&e, detJs[k]);
    }
    // CG
    cg = Eigen::Vector3d::Zero();
    for (auto n : e.nodes)
        cg += n->pos;
    cg /= e.nodes.size();
    // Memory is now valid and the volume can be computed
    valid = true;
    vol = e.Element::computeV();
}
