/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WMUMPS_H
#define WMUMPS_H

#include "tbox.h"
#ifdef WAVES_USE_MUMPS

#include "wLinearSolver.h"
#include <dmumps_c.h>
#include <Eigen/Sparse>
#include <limits>

namespace tbox
{

/**
 * @todo also output error in analyze and factorize, if any
 */

/**
 * @brief MUMPS interface
 * 
 * Reference: https://github.com/rboman/math0471/blob/master/mumps/main.cpp
 * Reference gmm_MUMPS_interface.h (old)
 * @authors Adrien Crovato
 */
class TBOX_API Mumps : public LinearSolver
{
private:
    DMUMPS_STRUC_C id;           ///< mumps structure
    std::vector<MUMPS_INT> rows; ///< rows id
    std::vector<MUMPS_INT> cols; ///< columns id
    std::vector<double> matx;    ///< matrix nonzero entries

    void configure(Eigen::SparseMatrix<double> const &A);
    void check();

public:
    Mumps();
    ~Mumps();

#ifndef SWIG
    virtual void analyze(Eigen::SparseMatrix<double> const &A) override;
    virtual void factorize(Eigen::SparseMatrix<double> const &A) override;
    virtual void solve(Eigen::Map<Eigen::VectorXd> const &b, Eigen::Map<Eigen::VectorXd> &x) override;
    virtual void compute(Eigen::SparseMatrix<double> const &A, Eigen::Map<Eigen::VectorXd> const &b, Eigen::Map<Eigen::VectorXd> &x) override;

    virtual double getError() override { return std::numeric_limits<double>::epsilon(); }
    virtual int getIterations() override { return 1; }

    virtual void write(std::ostream &out) const override;
#endif
    void setOption(int k, int v);
};

} // namespace tbox

#endif // WAVES_USE_MUMPS

#endif // WMUMPS_H
