/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WQUAD4_H
#define WQUAD4_H

#include "tbox.h"
#include "wLazy.h"
#include "wElement.h"
#include "wCacheQuad4.h"

namespace tbox
{

/**
 * @brief Linear quadrangle finite element
 * @authors Romain Boman, Adrien Crovato
 */
class TBOX_API Quad4 : public Element
{
#ifndef SWIG
    friend class MemQuad4;
    MemQuad4 *pmem; ///< private memory of precalculated values
    inline static CacheQuad4 &getCache();
    inline MemQuad4 &getMem();
    virtual double buildJ(size_t k, Eigen::MatrixXd &J, Eigen::MatrixXd &iJ) const override;
    virtual double buildJ(size_t k) const override;

    virtual Eigen::MatrixXd build(MATTYPE type) override;
    virtual Eigen::MatrixXd buildS() override;

    virtual double computeV(std::vector<double> const &u, Fct0 const &fct) override;
    virtual double computeV() override;
#endif

public:
    Quad4(int n, Tag *_ptag, Tag *_etag, std::vector<Tag *> &_parts, std::vector<Node *> &nods);
    virtual ~Quad4();
    virtual ELTYPE type() const override { return ELTYPE::QUAD4; }

#ifndef SWIG
    virtual Cache &getVCache() override;
    virtual Mem &getVMem() override;
    virtual Eigen::Vector3d normal() override;
#endif
};

#include "wQuad4.inl"

} // namespace tbox

#endif //WQUAD4_H
