/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wSfTetra4.h"

using namespace tbox;

void SfTetra4::eval(Eigen::Vector3d const &ksi, Eigen::VectorXd &ff) const
{
    ff.resize(4);
    double x = ksi(0);
    double y = ksi(1);
    double z = ksi(2);

    ff(0) = 1.0 - x - y - z;
    ff(1) = x;
    ff(2) = y;
    ff(3) = z;
}

void SfTetra4::evalD(Eigen::Vector3d const &ksi, Eigen::MatrixXd &dff) const
{
    dff.resize(3, 4);

    dff(0, 0) = -1.0;
    dff(1, 0) = -1.0;
    dff(2, 0) = -1.0;
    dff(0, 1) = 1.0;
    dff(1, 1) = 0.0;
    dff(2, 1) = 0.0;
    dff(0, 2) = 0.0;
    dff(1, 2) = 1.0;
    dff(2, 2) = 0.0;
    dff(0, 3) = 0.0;
    dff(1, 3) = 0.0;
    dff(2, 3) = 1.0;
}
